@@ -18,6 +18,7 @@ static char *prog;		      /* Program name */
 static int lwlport = Internet_Port + 2; /* Look Who's Listening port */
 static int debugging = FALSE;	      /* Debug mode enabled */
 static int verbose = FALSE;	      /* Show connections/disconnects */
+static int prolix = FALSE;	      /* Extremely verbose (show queries) */
 #ifdef HEXDUMP
 static int hexdump = FALSE;	      /* Dump received packets in hex ? */
 #endif
@@ -86,7 +87,7 @@ static char *dupSdesItem(cp)
   char **cp;
 {
     char *ip = *cp, *bp;
-    int l = ip[1];
+    int l = ip[1] & 0xFF;
 
     bp = malloc(l + 1);
     if (bp != NULL) {
@@ -405,15 +406,29 @@ dumpLwl(stderr, lf);
 	    lw = lw->next;
 	}
     }
-    updHTML();                        /* Update HTML is something's changed recently */
+    updHTML();                        /* Update HTML if something's changed recently */
+    signal(SIGALRM, release);	      /* Reset signal to handle timeout (Sys V) */
     alarm(TockTock);		      /* Reset the timer */
 }
 
+/*  PLUMBER  --  Catch SIGPIPE signal which occurs when remote user
+		 disconnects before we try to send the reply.  */
+
+static void plumber()
+{
+/*
+fprintf(stderr, "Caught SIGPIPE--continuing.\n");
+*/
+    signal(SIGPIPE, plumber);	      /* Reset signal just in case */
+}
+
 /*  EXITING  --  Release socket in case of termination.  */
 
 static void exiting()
 {
+/*
 fprintf(stderr, "Exiting.\n");
+*/
     shutdown(sock, 2);
     close(sock);
     exit(0);
@@ -424,7 +439,7 @@ fprintf(stderr, "Exiting.\n");
 static void usage()
 {
     V fprintf(stderr, "%s  --  Speak Freely: Look Who's Listening Server.\n", prog);
-    V fprintf(stderr, "               %s.\n", Release);
+    V fprintf(stderr, "               %s.\n", Relno);
     V fprintf(stderr, "\n");
     V fprintf(stderr, "Usage: %s [options]\n", prog);
     V fprintf(stderr, "Options:\n");
@@ -434,7 +449,7 @@ static void usage()
     V fprintf(stderr, "           -Mfile     Load server message from file\n");
     V fprintf(stderr, "           -Pport     Listen on given port\n");
     V fprintf(stderr, "           -U         Print this message\n");
-    V fprintf(stderr, "           -V         List connections and disconnects\n");
+    V fprintf(stderr, "           -V[V]      List connections and disconnects [-VV: all packets]\n");
 #ifdef HEXDUMP
     V fprintf(stderr, "           -X         Dump packets in hex\n");
 #endif
@@ -498,10 +513,15 @@ main(argc, argv)
 			if (message != NULL) {
 			    rtcp_t *rp = (rtcp_t *) message;
 
+#ifdef RationalWorld
 			    rp->common.version = RTP_VERSION;
 			    rp->common.p = 0;
 			    rp->common.count = 1;
 			    rp->common.pt = RTCP_APP;
+#else
+			    *((short *) rp) = htons((RTP_VERSION << 14) |
+						    RTCP_APP | (1 << 8));
+#endif
 			    rp->r.sdes.src = 0;
                             bcopy("SFmr", message + 8, 4);
 			    fread(message + 12, (int) fl, 1, fp);
@@ -522,6 +542,9 @@ main(argc, argv)
 
                 case 'V':             /*  -V  -- Show connects/disconnects */
 		    verbose = TRUE;
+                    if (op[1] == 'v' || op[1] == 'V') {
+			prolix = TRUE;
+		    }
 		    break;
 
 #ifdef HEXDUMP
@@ -549,10 +572,14 @@ main(argc, argv)
 	message = (char *) malloc(messagel);
 	rp = (rtcp_t *) message;
 
+#ifdef RationalWorld
 	rp->common.version = RTP_VERSION;
 	rp->common.p = 0;
 	rp->common.count = 1;
 	rp->common.pt = RTCP_APP;
+#else
+	*((short *) rp) = htons((RTP_VERSION << 14) | RTCP_APP | (1 << 8));
+#endif
 	rp->r.sdes.src = 0;
         bcopy("SFmr", message + 8, 4);
 	message[messagel - 1] = 0;
@@ -581,6 +608,8 @@ main(argc, argv)
     signal(SIGINT, exiting);	      /* Set signal to handle termination */
     signal(SIGTERM, exiting);	      /* Set signal to handle termination */
 
+    signal(SIGPIPE, plumber);         /* Catch "broken pipe" signals from disconnects */
+
     if (listen(sock, 5) < 0) {
         perror("calling listen for socket");
     }
@@ -604,11 +633,13 @@ main(argc, argv)
 
     while (TRUE) {
 	int csock, rll;
-	char p[1024];
-	rtcp_t *rp = (rtcp_t *) p;
+	char zp[1024];
+	rtcp_t *rp;
+	short srp;
 	struct sockaddr_in from;      /* Sending host address */
 	int fromlen;		      /* Length of sending host address */
 	struct lwl *lw; 	      /* Parsed packet */
+	char *p;
 
 	errno = 0;
 	do {
@@ -622,21 +653,59 @@ main(argc, argv)
             perror("accepting connection to socket");
 	    return 1;
 	}
+	if (prolix) {
+            printf("%s: %s accept\n", prog, inet_ntoa(from.sin_addr));
+	}
 
-	rll = recv(csock, p, sizeof p, 0);
+	rll = recv(csock, zp, sizeof zp, 0);
 #ifdef HEXDUMP
 	if (hexdump) {
             fprintf(stderr, "%s: %d bytes read from socket.\n", prog, rll);
-	    xd(stderr, p, rll, TRUE);
+	    xd(stderr, zp, rll, TRUE);
 	}
 #endif
 
-	/* Examine the packet to see what kind of request it is.  */
+	/* Walk through the individual items in a possibly composite
+           packet until we locate an item we're interested in.  This
+	   allows us to accept packets that comply with the RTP standard
+	   that all RTCP packets begin with an SR or RR.  */
 
-	if (rp->common.version == RTP_VERSION && /* Version ID correct */
-	    rp->common.count == 1) {
+	p = zp;
 
-	    switch (rp->common.pt) {
+	while ((p[0] >> 6 & 3) == RTP_VERSION) {
+	    int pt = p[1] & 0xFF;
+
+	    if (pt == RTCP_SDES || pt == RTCP_BYE || pt == RTCP_APP) {
+		break;
+	    }
+	    /* If not of interest to us, skip to next subpacket. */
+	    p += (ntohs(*((short *) (p + 2))) + 1) * 4;
+	}
+
+	/* Examine the packet to see what kind of request it is.
+           Note that since all the packets we're interested in can
+	   be parsed without knowing their actual length in bytes, we
+	   can ignore the possible presence of padding. */
+
+	rp = (rtcp_t *) p;
+	srp = ntohs(*((short *) p));
+	if (
+#ifdef RationalWorld
+	    rp->common.version == RTP_VERSION && /* Version ID correct */
+	    rp->common.count == 1
+#else
+	    (((srp >> 14) & 3) == RTP_VERSION) &&
+	    (((srp >> 8) & 0x1F) == 1)
+#endif
+	   ) {
+
+	    switch (
+#ifdef RationalWorld
+		    rp->common.pt
+#else
+		    srp & 0xFF
+#endif
+		   ) {
 
 		/*  SDES packet.  This is a notification by a listening
                                   that it's just started or is still
@@ -654,7 +723,7 @@ main(argc, argv)
 
 		    {
 			char *cp = (char *) rp->r.sdes.item,
-			     *lp = cp + (rp->common.length * 4);
+			     *lp = cp + (ntohs(rp->common.length) * 4);
 			struct lwl *lr, *llr;
 
 			lw = (struct lwl *) malloc(sizeof(struct lwl));
@@ -665,8 +734,12 @@ main(argc, argv)
 			    lw->naddr = from.sin_addr.s_addr;
 			    lw->port = from.sin_port;
 			    lw->ltime = time(NULL);
+			    if (prolix) {
+                                printf("%s: %s SDES %08X\n", prog,
+				    inet_ntoa(from.sin_addr), lw->ssrc);
+			    }
 			    while (cp < lp) {
-				switch (*cp) {
+				switch ((*cp) & 0xFF) {
 
 				    case RTCP_SDES_CNAME:
 					lw->cname = dupSdesItem(&cp);
@@ -785,6 +858,10 @@ main(argc, argv)
 		    {
 			struct lwl *lr = conn, *llr = NULL;
 
+			if (prolix) {
+                            printf("%s: %s BYE %08X\n", prog,
+				inet_ntoa(from.sin_addr), rp->r.bye.src[0]);
+			}
 			while (lr != NULL) {
 			    if (rp->r.bye.src[0] == lr->ssrc) {
 				if (llr == NULL) {
@@ -814,6 +891,10 @@ main(argc, argv)
 		    regarding the current state of the listening host list.  */
 
 		case RTCP_APP:
+		    if (prolix) {
+                        printf("%s: %s APP %.4s\n", prog,
+			    inet_ntoa(from.sin_addr), p + 8);
+		    }
 
 		    /*	SFlq  --  Pattern match in cname and name and
 				  return matches, up to the maximum
@@ -830,15 +911,20 @@ main(argc, argv)
 			char *ap;
 			int l, scandex = 0;
 
+#ifdef RationalWorld
 			rp->common.version = RTP_VERSION;
 			rp->common.p = 0;
 			rp->common.count = 0;
 			rp->common.pt = RTCP_SDES;
+#else
+			*((short *) rp) = htons((RTP_VERSION << 14) |
+						 RTCP_SDES | (0 << 8));
+#endif
 
 			ap = (char *) &(rp->r.sdes.src);
 
-			if (verbose) {
-                            printf("Query:   %s\n", p + 12);
+			if (prolix) {
+                            printf("%s: %s query \"%s\"\n", prog, inet_ntoa(from.sin_addr), p + 12);
 			}
 			lcase(p + 12);
 			while (lr != NULL) {
@@ -903,7 +989,14 @@ main(argc, argv)
 				while (((ap - b) & 3) != 0) {
 				    *ap++ = RTCP_SDES_END;
 				}
+#ifdef RationalWorld
 				rp->common.count++;
+#else
+				(((unsigned char *) rp)[0])++;
+#endif
+				if ((ap - b) > MaxReplyPacket) {
+				    break;
+				}
 			    }
 			    lr = lr->next;
 			    scandex++;
@@ -911,8 +1004,8 @@ main(argc, argv)
 
 			l = ap - b;
 
-			rp->common.length = ((l + 3) / 4) - 1;
-			l = (rp->common.length + 1) * 4;
+			rp->common.length = htons(((l + 3) / 4) - 1);
+			l = (ntohs(rp->common.length) + 1) * 4;
 #ifdef HEXDUMP
 			if (hexdump) {
                             fprintf(stderr, "%s: %d bytes sent to socket.\n", prog, l);
@@ -937,8 +1030,8 @@ main(argc, argv)
 			    xd(stderr, message, messagel, TRUE);
 			}
 #endif
-			if (verbose) {
-                            printf("Message request\n");
+			if (prolix) {
+                            printf("%s: %s server message request\n", prog, inet_ntoa(from.sin_addr));
 			}
 			if (send(csock, message, messagel, 0) < 0) {
                             perror("sending server message");
@@ -947,8 +1040,14 @@ main(argc, argv)
 		    break;
 
 		default:
-		    if (debugging) {
-                        fprintf(stderr, "Bogus payload type %d\n", rp->common.pt);
+		    if (debugging || verbose) {
+                        fprintf(stderr, "Bogus payload type %d\n",
+#ifdef RationalWorld
+				      rp->common.pt
+#else
+				      ((unsigned char *) rp)[1]
+#endif
+			       );
 		    }
 		    break;
 	    }
