@@ -12,18 +12,24 @@
 static int audiok = FALSE;	      /* Audio initialised flag */
 static int audiotime = 0;	      /* Audio timeout counter */
 static int debugging = FALSE;	      /* Debugging enabled */
-static int debugforce = 0;	      /* Debugging forced / prevented ? */
+static int debugforce = FALSE;	      /* Debugging forced / prevented ? */
+#ifdef HEXDUMP
+static int hexdump = FALSE;	      /* Dump received packets in hex ? */
+#endif
 static int noring = FALSE;	      /* Disable remote ring requests */
-static struct adpcm_state adpcm;      /* ADPCM compression state */
-static gsm gsmh;		      /* GSM handle */
+gsm gsmh;			      /* GSM handle */
 static char deskey[9] = "";           /* DES key, if any */
+static char rtpdeskey[9] = "";        /* RTP DES key, if any */
+static char vatdeskey[9] = "";        /* VAT DES key, if any */
 static char ideakey[17] = "";         /* IDEA key, if any */
 static char *curotp = NULL;	      /* Key file, if any */
 static int whichport = Internet_Port; /* Port to listen on (base address) */
-static int sock;		      /* Our socket */
+static int sock;		      /* Data socket */
+static int ssock;		      /* Control socket (RTP/VAT) */
 static struct sockaddr_in from;       /* Sending host address */
 static struct sockaddr_in name;       /* Address of destination host */
 static int fromlen;		      /* Length of sending host address */
+static int rll; 		      /* Length of packet read from socket */
 static int showhosts = FALSE;	      /* Show host names that connect */
 static int hosttimeout = 180 * 1000000L; /* Consider host idle after this time */
 static char *prog;		      /* Program name */ 
@@ -63,6 +69,8 @@ struct connection *conn = NULL;       /* Chain of current connections */
 
 static int timerStep = TockTock;      /* Current timer step */
 
+#define ucase(x)    (islower(x) ? toupper(x) : (x))
+
 /*  GSMDECOMP  --  Uncompress the contents of a sound buffer using GSM.  */
 
 static void gsmdecomp(sb)
@@ -89,45 +97,24 @@ static void gsmdecomp(sb)
     sb->buffer.buffer_len = declen;
 }
 
-/*  ADPCMDECOMP  --  Decompress the contents of a sound buffer using ADPCM.  */
-
-void adpcmdecomp(sb)
-  struct soundbuf *sb;
-{
-    char *dp = (char *) sb->buffer.buffer_val;
-    unsigned char *sp;
-    unsigned char dob[TINY_PACKETS * 2];
-
-    /* Restore the decoder state from the copy saved in the packet,
-       independent of the byte order of the machine we're running on. */
-
-    sp = (unsigned char *) dp + (sb->buffer.buffer_len - 3);
-    adpcm.valprev = (sp[0] << 8) | sp[1];
-    adpcm.index = sp[2];
-    sb->buffer.buffer_len -= 3;
-
-    adpcm_decoder_u(dp, dob, sb->buffer.buffer_len * 2, &adpcm);
-    sb->buffer.buffer_len *= 2;
-    bcopy(dob, dp, sb->buffer.buffer_len);
-}
-
 /*  LPCDECOMP  --  Uncompress the contents of a sound buffer using LPC.  */
 
-static void lpcdecomp(sb)
+static void lpcdecomp(sb, c)
   struct soundbuf *sb;
+  struct connection *c;
 {
     int i, j, l = 0;
     char *dpx = ((char *) sb->buffer.buffer_val) + sizeof(short);
     static char dcb[BUFL];
-    short declen = *((short *) sb->buffer.buffer_val);
+    short declen = ntohl(*((short *) sb->buffer.buffer_val));
 
-    if (declen <= 0 || declen > 1800) {
-	declen = 1800;
+    if (declen <= 0 || declen > (10 * LPC_FRAME_SIZE)) {
+	declen = 10 * LPC_FRAME_SIZE;
     }
     for (i = 0; l < declen;
-		i += sizeof(lpcparams_t)) {
-	j = lpc_synthesize((lpcparams_t *) (dpx + i), 1.0, dcb + l);
-	l += j;
+		i += LPCRECSIZE) {
+	lpc_synthesize(dcb + l, (lpcparams_t *) (dpx + i), &c->lpcc);
+	l += LPC_FRAME_SIZE;
     }
     bcopy(dcb, sb->buffer.buffer_val, declen);
     sb->buffer.buffer_len = declen;
@@ -165,10 +152,15 @@ static char *compressionType(msg)
 	   ((msg->compression & (fComp2X | fCompLPC)) ==
 				(fComp2X | fCompLPC)) ?
                                 "LPC+2X compressed" :
+	   ((msg->compression & (fComp2X | fCompVOX)) ==
+				(fComp2X | fCompVOX)) ?
+                                "VOX+2X compressed" :
            ((msg->compression & fCompADPCM) ? "ADPCM compressed" :
            ((msg->compression & fCompLPC) ? "LPC compressed" :
            ((msg->compression & fComp2X) ? "2X compressed" :
-           ((msg->compression & fCompGSM) ? "GSM compressed" :  "uncompressed"))));
+           ((msg->compression & fCompGSM) ? "GSM compressed" :
+           ((msg->compression & fCompVOX) ? "VOX compressed" :
+                                            "uncompressed")))));
 }
 
 /*  MAKESESSIONKEY  --	Generate session key.  */
@@ -261,7 +253,7 @@ static void sendLwlMessage(dobye)
 	    char v[1024];
 	    int l;
 
-            l = rtp_make_bye(v, ssrc, "Exiting sfspeaker");
+            l = rtp_make_bye(v, ssrc, "Exiting sfspeaker", FALSE);
 	    if (send(sock, v, l, 0) < 0) {
                 perror("sending look who's listening BYE packet");
 	    }
@@ -311,6 +303,8 @@ static void release()
 	    if (showhosts) {
                 fprintf(stderr, "%s: %s %s idle\n", prog, etime(), c->con_hostname);
 	    }
+	    c->con_protocol = PROTOCOL_UNKNOWN;
+	    c->con_uname[0] = c->con_email[0] = 0;
 
 	    /* If a face file transfer is in progress, shut it down and
 	       discard the incomplete face file. */
@@ -404,7 +398,7 @@ static void release()
 		c->face_timeout = 0;
 		bcopy((char *) &(c->con_addr), (char *) &(name.sin_addr),
 		    sizeof(struct in_addr));
-		fsb.compression = htonl(fFaceData | faceRequest);
+		fsb.compression = htonl(fProtocol | fFaceData | faceRequest);
 		fsb.buffer.buffer_len = htonl(c->face_address);
 		if (sendto(sock, (char *) &fsb,
 		    (int) (sizeof(struct soundbuf) - BUFL),
@@ -488,7 +482,15 @@ static void viewerterm()
 {
     int status, pid;
 
-    while ((pid = wait3(&status, WNOHANG, 0)) > 0) {
+    while ((pid = 
+#ifdef Solaris
+		  /* Should work for any System V Unix. */
+		  waitpid(0 /* or -1 */,&status, WNOHANG)
+#else
+		  /* BSD and IRIX implement wait3(). */
+		  wait3(&status, WNOHANG, 0)
+#endif
+	   ) > 0) {
 	if (WIFEXITED(status) || WIFSIGNALED(status)) {
 	    struct connection *c = conn;
 
@@ -543,18 +545,18 @@ static void playbuffer(msg, c)
     if (showhosts && (c->con_compmodes != (fCompressionModes &
 	    msg->compression))) {
 	c->con_compmodes = fCompressionModes & msg->compression;
-	if (showhosts) {
-            fprintf(stderr, "%s: %s sending %s.\n", prog, c->con_hostname,
-		    compressionType(msg));
-	}
-    }
-    if (Debug) {
-        fprintf(stderr, "%s: playing %d %s bytes from %s.\n", prog, msg->buffer.buffer_len,
-		compressionType(msg), c->con_hostname);
+        fprintf(stderr, "%s: %s sending %s.\n", prog, c->con_hostname,
+		compressionType(msg));
     }
+
     len = msg->buffer.buffer_len;
     val = msg->buffer.buffer_val;
 
+    if (Debug) {
+        fprintf(stderr, "%s: playing %d %s bytes from %s.\n", prog, len,
+		compressionType(msg), c->con_hostname);
+    }
+
     /* If the fSetDest bit is on, use the fDestJack bit to re-route
        the sound.  This is normally used to divert the sound to the
        speaker to get an individual's attention.  This can be
@@ -635,6 +637,11 @@ static void playbuffer(msg, c)
 
     /* If message is compressed, decompress appropriately. */
 
+    if (msg->compression & fCompVOX) {
+	vox_gsmdecomp(msg);
+	len = msg->buffer.buffer_len;
+    }
+
     if (msg->compression & fCompGSM) {
 	gsmdecomp(msg);
 	len = msg->buffer.buffer_len;
@@ -646,7 +653,7 @@ static void playbuffer(msg, c)
     }
 
     if (msg->compression & fCompLPC) {
-	lpcdecomp(msg);
+	lpcdecomp(msg, c);
 	len = msg->buffer.buffer_len;
     }
 
@@ -680,7 +687,7 @@ static void playbuffer(msg, c)
 static void usage()
 {
     V fprintf(stderr, "%s  --  Speak Freely sound receiver.\n", prog);
-    V fprintf(stderr, "               %s.\n", Release);
+    V fprintf(stderr, "               %s.\n", Relno);
     V fprintf(stderr, "\n");
     V fprintf(stderr, "Usage: %s [options]\n", prog);
     V fprintf(stderr, "Options:\n");
@@ -698,6 +705,9 @@ static void usage()
     V fprintf(stderr, "           -R[+]file        Record [append] sound in file\n");
     V fprintf(stderr, "           -U               Print this message\n");
     V fprintf(stderr, "           -Vtimeout        Show hostnames that connect\n");
+#ifdef HEXDUMP
+    V fprintf(stderr, "           -X               Dump packets in hex\n");
+#endif
     V fprintf(stderr, "           -Z\"phrase\"       Set PGP secret key pass phrase\n");
     V fprintf(stderr, "\n");
     V fprintf(stderr, "by John Walker\n");
@@ -716,17 +726,81 @@ main(argc, argv)
     FILE *fp;
     struct connection *c;
     char *cp;
-    int newconn;
+    int newconn, wasrtp;
     struct auhdr {		      /* .au file header */
 	char magic[4];
 	long hsize, dsize, emode, rate, nchan;
     };
     static struct auhdr afh = { ".snd", sizeof(struct auhdr), 1, 1, 8000, 1 };
 
+    prog = argv[0];
+
+    /* First pass option processing.  We have to first scan
+       the options to handle any which affect creation of the
+       socket.	One the second pass we can assume the socket
+       already exists, allowing us to join multicast groups,
+       etc. */
+
+    for (i = 1; i < argc; i++) {
+	char *op, opt;
+
+	op = argv[i];
+        if (*op == '-') {
+	    opt = *(++op);
+	    if (islower(opt)) {
+		opt = toupper(opt);
+	    }
+
+	    switch (opt) {
+
+                case 'P':             /* -Pport  --  Port to listen on */
+		    whichport = atoi(op + 1);
+		    break;
+
+                case 'U':             /* -U  --  Print usage information */
+                case '?':             /* -?  --  Print usage information */
+		    usage();
+		    return 0;
+	    }
+	} else {
+	    usage();
+	    return 2;
+	}
+    }
+
+    /* Create the sockets from which to read */
+
+    sock = socket(AF_INET, SOCK_DGRAM, 0);
+    if (sock < 0) {
+        perror("opening data socket");
+	return 1;
+    }
+
+    ssock = socket(AF_INET, SOCK_DGRAM, 0);
+    if (ssock < 0) {
+        perror("opening control socket");
+	return 1;
+    }
+
+    /* Create name with wildcards. */
+
+    name.sin_family = AF_INET;
+    name.sin_addr.s_addr = INADDR_ANY;
+    name.sin_port = htons(whichport);
+    if (bind(sock, (struct sockaddr *) &name, sizeof name) < 0) {
+        perror("binding data socket");
+	return 1;
+    }
+    name.sin_port = htons(whichport + 1);
+    if (bind(ssock, (struct sockaddr *) &name, sizeof name) < 0) {
+        perror("binding control socket");
+	return 1;
+    }
+    name.sin_port = htons(whichport);
+
     /*	Process command line options.  */
 
     pgppass = getenv("PGPPASS");
-    prog = argv[0];
 
 #if Internet_Port != 2074
     /* This is a reminder to the author who frequently tests with
@@ -761,7 +835,7 @@ main(argc, argv)
 		    break;
 
                 case 'D':             /* -D  --  Force debug output */
-		    debugforce = 1;
+		    debugforce = TRUE;
 		    break;
 
                 case 'I':             /* -Ikey  --  Set IDEA key */
@@ -780,10 +854,10 @@ main(argc, argv)
                 case 'K':             /* -Kkey  --  Set DES key */
 		    desinit(1);       /* Initialise the DES library */
 		    if (strlen(op + 1) == 0) {
-			deskey[0] = FALSE;
+			deskey[0] = rtpdeskey[0] = vatdeskey[0] = FALSE;
 		    } else {
 			struct MD5Context md5c;
-			char md5key[16];
+			char md5key[16], algorithm[16];
 
 			MD5Init(&md5c);
 			MD5Update(&md5c, op + 1, strlen(op + 1));
@@ -793,6 +867,14 @@ main(argc, argv)
 					  ((md5key[j] ^ md5key[j + 8]) & 0x7F);
 			}
 			deskey[0] = TRUE;
+			des_string_to_key(op + 1, (des_cblock *) (vatdeskey + 1));
+			string_DES_key(op + 1, rtpdeskey + 1, algorithm);
+                        if (strcmp(algorithm, "DES-CBC") != 0) {
+                            fprintf(stderr, "Unsupported encryption algorithm: %s.  Only DES-CBC is available.\n",
+				    algorithm);
+			    return 2;
+			}
+			rtpdeskey[0] = vatdeskey[0] = TRUE;
 		    }
 		    break;
 
@@ -861,10 +943,6 @@ main(argc, argv)
 		    }
 		    break;
 
-                case 'P':             /* -Pport  --  Port to listen on */
-		    whichport = atoi(op + 1);
-		    break;
-
                 case 'Q':             /* -Q  --  Prevent debug output */
 		    debugforce = -1;
 		    break;
@@ -891,12 +969,7 @@ main(argc, argv)
 		    }
 		    break;
 
-                case 'U':             /* -U  --  Print usage information */
-                case '?':             /* -?  --  Print usage information */
-		    usage();
-		    return 0;
-
-                case 'V':             /* -V  --  Show hostnames that connect */
+                case 'V':             /* -Vtimeout  --  Show hostnames that connect */
 		    showhosts = TRUE;
 		    if (op[1] != 0) {
 			int t = atoi(op + 1) * 1000000L;
@@ -910,6 +983,12 @@ main(argc, argv)
 		    }
 		    break;
 
+#ifdef HEXDUMP
+                case 'X':             /* -X  --  Dump packets in hex */
+		    hexdump = debugforce = TRUE;
+		    break;
+#endif
+
                 case 'Z':             /* -Z"phrase"  --  Set PGP pass phrase */
 		    if (op[1] != 0) {
 			pgppass = strdup(op + 1);
@@ -933,30 +1012,9 @@ main(argc, argv)
 		    }
 		    break;
 	    }
-	} else {
-	    usage();
-	    return 2;
 	}
     }
 
-    /* Create the socket from which to read */
-
-    sock = socket(AF_INET, SOCK_DGRAM, 0);
-    if (sock < 0) {
-        perror("opening datagram socket");
-	return 1;
-    }
-
-    /* Create name with wildcards. */
-
-    name.sin_family = AF_INET;
-    name.sin_addr.s_addr = INADDR_ANY;
-    name.sin_port = htons(whichport);
-    if (bind(sock, (struct sockaddr *) &name, sizeof name) < 0) {
-        perror("binding datagram socket");
-	return 1;
-    }
-
 #ifdef HALF_DUPLEX
     {
 	struct hostent *h;
@@ -981,7 +1039,10 @@ main(argc, argv)
 
     /* Initialise LPC decoding. */
 
-    lpc_init(180);
+    if (!lpc_start()) {
+        fprintf(stderr, "Cannot allocate LPC decoding memory.\n");
+	return 1;
+    }
 
     /* Find assigned port value and print it. */
 
@@ -1016,7 +1077,7 @@ main(argc, argv)
 	    makeSessionKey(md5key);
 	    bcopy(md5key, (char *) &ssrc, sizeof ssrc);
 
-	    sdesl = rtp_make_sdes(&sdes, ssrc, whichport);
+	    sdesl = rtp_make_sdes(&sdes, ssrc, whichport, FALSE);
 	    sendLwlMessage(FALSE);
 	} else {
             fprintf(stderr, "%s: warning, SPEAKFREE_LWL_TELL host %s unknown.\n",
@@ -1047,20 +1108,34 @@ main(argc, argv)
     /* Read from the socket. */
 
     while (TRUE) {
-	int rll;
+	fd_set fdset;
+	int wsock, control;
 
-	fromlen = sizeof(from);
-	if ((rll = recvfrom(sock, (char *) &sb, sizeof sb, 0, (struct sockaddr *) &from, &fromlen)) < 0) {
-#ifdef sgi
-	    if (errno != EINTR) {
-                perror("receiving datagram packet");
-	    }
-#endif
+	FD_ZERO(&fdset);
+	FD_SET(sock, &fdset);
+	FD_SET(ssock, &fdset);
+	if (select(FD_SETSIZE, &fdset, NULL, NULL, NULL) <= 0) {
 	    continue;
 	}
-#ifdef SHOW_SOCKET
-	if (Debug) {
-            fprintf(stderr, "%s: %d bytes read from socket.\n", prog, rll);
+	wsock = FD_ISSET(sock, &fdset) ? sock :
+		    (FD_ISSET(ssock, &fdset) ? ssock : -1);
+	if (wsock < 0) {
+	    continue;
+	}
+	control = wsock == ssock;
+	fromlen = sizeof(from);
+	if ((rll = recvfrom(wsock, (char *) &sb, sizeof sb, 0, (struct sockaddr *) &from, &fromlen)) < 0) {
+	    if (errno != EINTR) {
+                perror(!control ? "receiving data packet" :
+                                  "receiving control packet");
+	    }
+	    continue;
+	}
+#ifdef HEXDUMP
+	if (hexdump) {
+            fprintf(stderr, "%s: %d bytes read from %s socket.\n", prog, rll,
+                    control ? "control" : "data", rll);
+	    xd(stderr, &sb, rll, TRUE);
 	}
 #endif
 
@@ -1085,6 +1160,7 @@ main(argc, argv)
 		c->con_next = conn;
 		c->pgpkey[0] = FALSE;
 		bzero(c->keymd5, 16);
+		c->con_uname[0] = c->con_email[0] = 0;
 		conn = c;
 		bcopy(&from.sin_addr, &(c->con_addr),
 		    sizeof(struct in_addr));
@@ -1112,6 +1188,10 @@ main(argc, argv)
 	    c->face_address = 0L;
 	    c->face_retry = 0;
 	    c->con_compmodes = -1;
+	    c->con_protocol = PROTOCOL_UNKNOWN;
+	    c->con_reply_current = FALSE;
+            bcopy("\221\007\311\201", c->con_session, 4);
+	    lpc_init(&c->lpcc);
 	}
 
 	if (c != NULL) {
@@ -1121,24 +1201,6 @@ main(argc, argv)
 		if (showhosts) {
                     fprintf(stderr, "%s: %s %s connect\n", prog, etime(), c->con_hostname);
 		}
-
-		/* If a reply file is requested, create an executable
-		   shell script to reply to the host that just connected
-		   using either the default command or whatever command
-		   the user specifies on the -A option. */
-
-		if (replyfile != NULL) {
-                    FILE *rfp = fopen(replyfile, "w");
-
-		    if (rfp != NULL) {
-                        fprintf(rfp, "#! /bin/sh\n");
-			/* $* allows specifying options on reply file call. */
-                        fprintf(rfp, "%s $* %s\n", replycmd, c->con_hostname);
-			/* Make the file executable. */
-			fchmod(fileno(rfp), 0755);
-			fclose(rfp);
-		    }
-		}
 	    }
 
 	    /* Request face data from the other end, starting with
@@ -1151,7 +1213,8 @@ main(argc, argv)
                end, which didn't trigger a reciprocal request by
 	       us. */
 
-	    if ((c->face_stat == FSinit) &&
+	    if (!control && (c->con_protocol == PROTOCOL_SPEAKFREE) &&
+		    (c->face_stat == FSinit) &&
 		    isSoundPacket(ntohl(sb.compression)) &&
 		    (ntohl(sb.compression) & fFaceOffer)) {
 		c->face_address = 0;
@@ -1168,27 +1231,380 @@ main(argc, argv)
 	    continue;
 	}
 
-#ifdef RTP_SUPPORT
+	wasrtp = FALSE;
+
+        /* If a DES key is present and we're talking RTP or VAT
+	   protocol we must decrypt the packet at this point.
+	   We decrypt the packet if:
+
+		1.  A DES key was given on the command line, and
+		    either:
+
+		    a)	The packet arrived on the control port
+			(and hence must be from an RTP/VAT client), or
+
+		    b)	The protocol has already been detected as
+			RTP or VAT by reception of a valid control
+			port message.  */
+
+	if ((control || (c->con_protocol == PROTOCOL_RTP) ||
+			(c->con_protocol == PROTOCOL_VAT)) &&
+	     rtpdeskey[0]) {
+
+	    /* One more little twist.  If this packet arrived on the
+	       control channel, see if it passes all the tests for a
+               valid RTCP packet.  If so, we'll assume it isn't
+	       encrypted.  RTP utilities have the option of either
+	       encrypting their control packets or sending them in
+	       the clear, so a hack like this is the only way we have
+	       to guess whether something we receive is encrypted. */
+
+	    if (!isValidRTCPpacket((unsigned char *) &sb, rll)) {
+		des_cblock key;
+		des_key_schedule sched;
+		des_cblock ivec;
+		int drll = rll;
+		char *whichkey;
+		static int toggle = 0;
+
+		bzero(ivec, 8);
+		drll = (rll + 7) & (~7);
+		if (Debug) {
+                    fprintf(stderr, "Decrypting %d VAT/RTP bytes with DES key.\r\n",
+			    drll);
+		}
+		if (drll > rll) {
+		    /* Should only happen for VAT protocol.  Zero the rest of
+		       the DES encryption block to guarantee consistency. */
+		    bzero(((char *) &sb) + rll, drll - rll);
+		}
+
+		/* If the protocol is unknown, toggle back and forth
+		   between the RTP and VAT DES keys until we crack the
+		   packet and set the protocol. */
+
+		if (c->con_protocol == PROTOCOL_UNKNOWN || 
+		    c->con_protocol == PROTOCOL_VATRTP_CRYPT ||
+		    c->con_protocol == PROTOCOL_SPEAKFREE) {
+		    whichkey = toggle == 0 ? vatdeskey :
+			       (toggle == 1 ? rtpdeskey : NULL);
+		    toggle = (toggle + 1) % 3;
+		} else {
+		    whichkey = c->con_protocol == PROTOCOL_VAT ?
+				vatdeskey : rtpdeskey;
+		}
+		if (whichkey != NULL) {
+		    des_set_key((des_cblock *) (whichkey + 1), sched);
+		    des_ncbc_encrypt((des_cblock *) &sb,
+			(des_cblock *) &sb, rll, sched,
+			(des_cblock *) ivec, DES_DECRYPT);
+
+		    /* Just one more thing.  In RTP (unlike VAT), when
+		       an RTCP control packet is encrypted, 4 bytes of
+		       random data are prefixed to the packet to prevent
+		       known-plaintext attacks.  We have to strip this
+		       prefix after decrypting. */
+
+		    if (control && ((*(((char *) &sb) + 4) & 0xC0) == 0x80)) {
+			rll -= 4;
+			bcopy(((char *) &sb) + 4, (char *) &sb, rll);
+		    }
+		}
+#ifdef HEXDUMP
+		if (hexdump) {
+		    xd(stderr, &sb, rll, TRUE);
+		}
+#endif
+	    }
+	}
+
+	/* If this packet arrived on the session control port, dispatch
+	   it to the appropriate handler for its protocol. */
+
+	if (control) {
+	    short protocol = PROTOCOL_VATRTP_CRYPT;
+	    unsigned char *p = (unsigned char *) &sb;
+	    int proto = (p[0] >> 6) & 3;
+
+	    if (proto == 0) {
+		/* To avoid spoofing by bad encryption keys, require
+                   a proper ID message be seen before we'll flip into
+		   VAT protocol. */
+		if (((p[1] == 1) || (p[1] == 3)) ||
+		    ((c->con_protocol == PROTOCOL_VAT) && (p[1] == 2))) {
+		    protocol = PROTOCOL_VAT;
+		    bcopy(p + 2, c->con_session, 2);  /* Save conference ID */
+
+		    if (p[1] == 1 && showhosts) {
+			char uname[256];
+
+			bcopy(p + 4, uname, rll - 4);
+			uname[rll - 4] = 0;
+			if (strcmp(uname, c->con_uname) != 0) {
+			    strcpy(c->con_uname, uname);
+                            fprintf(stderr, "%s: %s sending from %s.\n", prog,
+				    c->con_uname, c->con_hostname);
+			}
+		    }
+
+		    /* Handling of VAT IDLIST could be a lot more elegant
+		       than this. */
+
+		    if (p[1] == 3 && showhosts) {
+			char *uname;
+
+			uname = (char *) malloc(rll);
+			if (uname != NULL) {
+			    unsigned char *bp = p, *ep = p + rll;
+			    int i = bp[4];
+
+			    bp += 8;
+			    uname[0] = 0;
+			    *ep = 0;
+			    while (--i >= 0 && bp < ep) {
+				unsigned long id = *(unsigned long *) bp;
+
+				bp += 4;
+                                strcat(uname, "\t");
+				strcat(uname, (char *) bp);
+				while (isspace(uname[strlen(uname) - 1])) {
+				    uname[strlen(uname) - 1] = 0;
+				}
+                                strcat(uname, "\n");
+				bp += (strlen((char *) bp) + 3) & ~3;
+			    }
+			    if (strncmp(uname, c->con_uname, (sizeof c->con_uname - 1)) != 0) {
+				strncpy(c->con_uname, uname, sizeof c->con_uname);
+				if (strlen(uname) > ((sizeof c->con_uname) - 1)) {
+				    c->con_uname[((sizeof c->con_uname) - 1)] = 0;
+				}
+                                fprintf(stderr, "%s: now in conference at %s:\n%s", prog,
+					c->con_hostname, uname);
+			    }
+			    free(uname);
+			}
+		    }
+
+                    /* If it's a DONE packet, reset protocol to unknown. */
+
+		    if (p[1] == 2) {
+			c->con_protocol = protocol = PROTOCOL_UNKNOWN;
+			c->con_uname[0] = c->con_email[0] = 0;
+			if (showhosts) {
+                            fprintf(stderr, "%s: %s VAT connection closed.\n",
+				    prog, c->con_hostname);
+			}
+		    }
+		}
+	    } else if (proto == RTP_VERSION || proto == 1) {
+		if (isValidRTCPpacket((unsigned char *) &sb, rll)) {
+		    protocol = (proto == 1) ? PROTOCOL_SPEAKFREE : PROTOCOL_RTP;
+		    bcopy(p + 4, c->con_session, 4);  /* Save SSRC */
+                    /* If it's a BYE packet, reset protocol to unknown. */
+		    if (isRTCPByepacket((unsigned char *) &sb, rll)) {
+			c->con_protocol = protocol = PROTOCOL_UNKNOWN;
+			c->con_uname[0] = c->con_email[0] = 0;
+			if (showhosts) {
+                            fprintf(stderr, "%s: %s %s connection closed.\n",
+				    prog, c->con_hostname,
+                                    proto == 1 ? "Speak Freely" : "RTP");
+			}
+		    } else if (showhosts) {
+			struct rtcp_sdes_request rp;
+
+			rp.nitems = 4;
+			rp.item[0].r_item = RTCP_SDES_CNAME;
+			rp.item[1].r_item = RTCP_SDES_NAME;
+			rp.item[2].r_item = RTCP_SDES_EMAIL;
+			rp.item[3].r_item = RTCP_SDES_TOOL;
+			if (parseSDES((unsigned char *) &sb, &rp)) {
+			    char uname[256], email[256];
+
+			    uname[0] = email[0] = 0;
+			    if (rp.item[1].r_text != NULL) {
+				copySDESitem(rp.item[1].r_text, uname);
+				if (rp.item[2].r_text != NULL) {
+				    copySDESitem(rp.item[2].r_text, email);
+				} else if (rp.item[2].r_text != NULL) {
+				    copySDESitem(rp.item[0].r_text, email);
+				}
+			    } else if (rp.item[2].r_text != NULL) {
+				copySDESitem(rp.item[2].r_text, uname);
+			    } else if (rp.item[0].r_text != NULL) {
+				copySDESitem(rp.item[0].r_text, uname);
+			    }
+			    if (strcmp(uname, c->con_uname) != 0 ||
+				strcmp(email, c->con_email) != 0) {
+				strcpy(c->con_uname, uname);
+				strcpy(c->con_email, email);
+				if (uname[0]) {
+                                    fprintf(stderr, "%s: %s", prog, uname);
+				    if (email[0]) {
+                                      fprintf(stderr, " (%s)", email);
+				    }
+                                    fprintf(stderr, " sending from %s.\n",
+					    c->con_hostname);
+				}
+			    }
+			}
+		    }
+		} else {
+		    if (Debug) {
+                        fprintf(stderr, "Invalid RTCP packet received.\n");
+		    }
+		}
+	    } else {
+		if (Debug) {
+                    fprintf(stderr, "Bogus protocol 3 control message.\n");
+		}
+	    }
+
+	    /* If protocol changed, update in connection and, if appropriate,
+	       update the reply command. */
+
+	    if (protocol != c->con_protocol) {
+		static char *pname[] = {
+                    "Speak Freely",
+                    "VAT",
+                    "RTP",
+                    "VAT/RTP encrypted",
+                    "Unknown"
+		};
+
+		c->con_protocol = protocol;
+		if (showhosts) {
+                    fprintf(stderr, "%s: %s sending in %s protocol.\n",
+			    prog, c->con_hostname, pname[protocol]);
+		}
+		c->con_reply_current = FALSE;
+	    }
+	    continue;
+	}
+
+	/* If this message is tagged with our Speak Freely protocol
+	   bit, force protocol back to Speak Freely.  This allows us
+	   to switch back to Speak Freely after receiving packets in
+	   VAT.  We can still get confused if we receive a packet from
+           an older version of Speak Freely that doesn't tag. */
+
+	if (c->con_protocol == PROTOCOL_VAT ||
+	    c->con_protocol == PROTOCOL_VATRTP_CRYPT) {
+	    unsigned char *p = (unsigned char *) &sb;
+
+	    if (((p[0] >> 6) & 3) == 1) {
+		c->con_protocol = PROTOCOL_SPEAKFREE;
+	    }
+	}
+
+	/* If this is a VAT packet, translate it into a sound buffer. */
+
+	if (((c->con_protocol == PROTOCOL_VAT)) &&
+	    (bcmp(((unsigned char *) &sb) + 2, c->con_session, 2) == 0) &&
+	    isvat((unsigned char *) &sb, rll)) {
+	    if (sb.buffer.buffer_len == 0) {
+		if (Debug) {
+                    fprintf(stderr, "Ignoring unparseable VAT packet.\n");
+		}
+		continue;
+	    }
+	    wasrtp = TRUE;
 
 	/* If this is an RTP packet, transmogrify it into a sound
 	   buffer we can understand. */
 
-	if (isrtp((unsigned char *) &sb, rll, c)) {
+	} else if ((c->con_protocol == PROTOCOL_RTP) &&
+		 (bcmp(((unsigned char *) &sb) + 8, c->con_session, 4) == 0) &&
+		 isrtp((unsigned char *) &sb, rll)) {
 	    if (sb.buffer.buffer_len == 0) {
-                fprintf(stderr, "Ignoring unparseable RTP packet.\n");
+		if (Debug) {
+                    fprintf(stderr, "Ignoring unparseable RTP packet.\n");
+		}
 		continue;
 	    }
-	} else {
-#endif
+	    wasrtp = TRUE;
+	}
+
+	if (!wasrtp) {
+	    int xbl;
 
 	    /* Convert relevant fields from network to host
 	       byte order, if necessary. */
 
 	    sb.compression = ntohl(sb.compression);
 	    sb.buffer.buffer_len = ntohl(sb.buffer.buffer_len);
-#ifdef RTP_SUPPORT
+
+	    /* Now if this is a valid Speak Freely packet (as
+	       opposed to a VAT packet masquerading as one, or
+               an encrypted VAT or RTP packet we don't have the
+	       proper key to decode), the length received from the
+	       socket will exactly equal the buffer length plus
+	       the size of the header.	This is a reasonably
+	       good validity check, well worth it considering the
+	       horrors treating undecipherable garbage as a sound
+	       buffer could inflict on us. */
+
+	    xbl = sb.buffer.buffer_len + (sizeof(struct soundbuf) - BUFL);
+	    if ((sb.compression & (fEncDES | fEncIDEA | fEncPGP)) != 0) {
+		xbl = ((sb.buffer.buffer_len + 7) & (~7)) +
+		      (sizeof(struct soundbuf) - BUFL);
+	    }
+	    if (xbl != rll) {
+		if (Debug) {
+		    fprintf(stderr,
+                            "Sound buffer length %ld doesn't match %ld byte packet.  Hdr=%08X\n",
+			    xbl, rll, sb.compression);
+		}
+		if (showhosts && c->con_protocol != PROTOCOL_UNKNOWN) {
+                    fprintf(stderr, "%s: %s sending in unknown protocol or encryption.\n",
+			    prog, c->con_hostname);
+		}
+		c->con_protocol = PROTOCOL_UNKNOWN;
+		continue;
+	    }
+
+	    /* It does appear to be a genuine Speak Freely sound
+	       buffer.	On that basis, set the protocol to Speak Freely
+               even if the buffer isn't explicitly tagged. */
+
+	    if (c->con_protocol != PROTOCOL_SPEAKFREE) {
+		c->con_protocol = PROTOCOL_SPEAKFREE;
+		if (showhosts) {
+                    fprintf(stderr, "%s: %s sending in Speak Freely protocol.\n", prog, c->con_hostname);
+		}
+		c->con_reply_current = FALSE;
+	    }
+	}
+
+        /* If this is the first sound packet we've seen from this host 
+	   in this protocol, create (or update) the reply command file. */
+
+	if (replyfile != NULL && !c->con_reply_current) {
+
+	    /* If a reply file is requested, create an executable
+	       shell script to reply to the host that just connected
+	       using either the default command or whatever command
+	       the user specifies on the -A option. */
+
+            FILE *rfp = fopen(replyfile, "w");
+	    static char *popt[] = {
+                "",
+                "-vat ",
+                "-rtp ",
+                "-rtp ",
+                ""
+	    };
+
+	    if (rfp != NULL) {
+                fprintf(rfp, "#! /bin/sh\n");
+		/* $* allows specifying options on reply file call. */
+                fprintf(rfp, "%s %s$* %s\n", replycmd,
+			      popt[c->con_protocol], c->con_hostname);
+		/* Make the file executable. */
+		fchmod(fileno(rfp), 0755);
+		fclose(rfp);
+	    }
+	    c->con_reply_current = TRUE;
 	}
-#endif
 
 #ifdef HALF_DUPLEX
 
@@ -1241,7 +1657,7 @@ main(argc, argv)
 		    *((long *) sb.buffer.buffer_val) = htonl(sb.buffer.buffer_len);
 		    l = fread(sb.buffer.buffer_val + sizeof(long),
 			1, 512 - (sizeof(long) + (sizeof(soundbuf) - BUFL)), facefile);
-		    sb.compression = fFaceData | faceReply;
+		    sb.compression = fProtocol | fFaceData | faceReply;
 		    if (Debug) {
                         fprintf(stderr, "%s: sending %d bytes of face data at %ld to %s\n",
 			    prog, l, ntohl(*((long *) sb.buffer.buffer_val)), c->con_hostname);
@@ -1249,7 +1665,7 @@ main(argc, argv)
 		    l += sizeof(long);
 		} else {
 		    /* No face file.  Shut down requestor. */
-		    sb.compression = fFaceData | faceLess;
+		    sb.compression = fProtocol | fFaceData | faceLess;
 		    l = 0;
 		}
 		bcopy((char *) &(from.sin_addr), (char *) &(name.sin_addr),
@@ -1460,7 +1876,7 @@ main(argc, argv)
 			}
 		    }
 #endif
-                    sprintf(cmd, "pgp -f +nomanual +verbose=0 %s%s%s <%s",
+                    sprintf(cmd, "pgp -f +nomanual +verbose=0 +armor=off %s%s%s <%s",
                         pgppass ? "-z\"" : "", pgppass ? pgppass : "",
                         pgppass ? "\" " : "", f);
 #ifdef PGP_DEBUG
@@ -1474,7 +1890,24 @@ main(argc, argv)
 		    if (pipe == NULL) {
                         fprintf(stderr, "Unable to open pipe to: %s\n", c);
 		    } else {
-			if (fread(c->pgpkey, 1, 17, pipe) == 17) {
+			int lr;
+
+			/* Okay, explanation time again.  On some systems
+			   (Silicon Graphics, for example), the timer tick
+			   alarm signal can cause the pending read from the
+                           PGP key pipe to return an "Interrupted system
+                           call" status (EINTR) with (as far as I've ever
+                           seen and I sincerely hope it's always) zero bytes
+			   read.  This happens frequently when the timer is
+			   running and the user takes longer to enter the
+			   secret key pass phrase than the timer tick.	So,
+			   if this happens we keep on re-issuing the pipe
+			   read until the phrase allows PGP to finish the
+			   job. */
+
+			while ((lr = fread(c->pgpkey, 1, 17, pipe)) != 17 &&
+			       (errno == EINTR)) ;
+			if (lr == 17) {
 			    c->pgpkey[0] = TRUE;
 #ifdef PGP_DEBUG
 			    {	
@@ -1494,6 +1927,10 @@ main(argc, argv)
 			} else {
 			    c->pgpkey[0] = FALSE;
                             fprintf(stderr, "%s: Error decoding PGP session key.\n", prog);
+#ifdef PGP_DEBUG
+                            fprintf(stderr, "Read status from pipe: %d\n", lr);
+                            perror("reading decoded PGP key from pipe");
+#endif
 			}
 			pclose(pipe);
 		    }
@@ -1508,6 +1945,7 @@ main(argc, argv)
     close(sock);
     desdone();
     gsm_destroy(gsmh);
+    lpc_end();
     exiting();
     if (record != NULL) {
 	fclose(record);
